//
// convert.c -- Read in the file, convert it to internal format, write it out
//
// These functions should be usable on both the Windows version and the
// console version.
//
// Copyright (C) 2000 by Bob Crispen <crispen@hiwaay.net> doing business as
//   Foxy Software
//
// LICENSE:
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stdio.h>
#include <string.h>
#include "vrml.h"
#include "convert.h"
#include "orient.h"
#ifdef CONSOLE_VERSION
    #include "consolefuncs.h"
    #ifdef MSDOS
	#define WRITEMODE "wb"
    #else
	#define WRITEMODE "w"
    #endif
#else
    #define WRITEMODE "wb"
    #include "winfuncs.h"
#endif

#define BUFLEN 1000

FILE* infile;
char buf[BUFLEN];
char* tokptr;

// Read the next token from the input file
// Returns: pointer to the next token in the input stream or NULL if EOF
char* next_token()
{
    int bufptr = 0;
    char* token;

    while ((token = strtok(tokptr, " \t\0")) == NULL) {
	memset(buf, 0, BUFLEN);
	if (fgets(buf, BUFLEN, infile) == NULL) {
	    fclose(infile);
	    return NULL;
	}
	// Clean out CR and LF
	bufptr = strlen(buf)-1;
	for (;(buf[bufptr] == '\r') || (buf[bufptr] == '\n'); --bufptr)
	    buf[bufptr] = '\0';
	tokptr = buf;
    }
    tokptr = NULL;
    // Comment character skips to end of line
    if (token[0] == '#') {
	memset(buf, 0, BUFLEN);
	tokptr = buf;
	return next_token();
    }
    return token;
}


// Read the file and convert if from 3DMF into our internal format
int read_in(char *fileName)
{

// State machine enums
#define SCANNING 0
#define READING_TRIMESH_SPEC 1
#define READING_TRIANGLES 2
#define READING_EDGES 3
#define READING_POINTS 4
#define SEARCHING_FOR_DIFFUSECOLOR 5
#define READING_DIFFUSECOLOR 6
#define READING_FORWARDDIRECTION 7
#define READING_CAMERAPLACEMENT 8
#define READING_DIRECTIONALLIGHT 9
#define SEARCHING_FOR_LIGHT 10
#define READING_LIGHT 11
#define READING_AMBIENTINTENSITY 12

    char lastDefName[BUFLEN];
    char outputLine[BUFLEN];
    int meshParms[7];
#define NUM_TRIANGLES meshParms[1]
#define NUM_EDGES meshParms[3]
#define NUM_POINTS meshParms[5]
    int count;
    char *token;
    int state = SCANNING;
    float coord;
    int index;
    int direction[3];
    float lightDirection[3];
    float color;
    float intensity;
    struct Shape* sp;
    struct Viewpoint* vp;
    struct DirectionalLight* lp;
    Vector pos, lookat, up;
    Vector axis;
    double angle;
    float ambientIntensity = 0.0;
    long int filePos;
    float fileLength;
    int pct;

    if (!strcmp(fileName, "stdin")) {
	infile = stdin;
	fileLength = 100;
    } else {
	if ((infile = fopen(fileName, "rb")) == NULL)
	    return CNV_CANT_OPEN;
	// Find the file length
	filePos = ftell(infile);
	fseek(infile, 0, SEEK_END);
	fileLength = (float)ftell(infile);
	fseek(infile, filePos, SEEK_SET);
    }

    // Just in case another file has been read in already
    clean_up();

    memset(buf, 0, BUFLEN);
    memset(lastDefName, 0, BUFLEN);

    // Set up to read the first token
    if (fgets(buf, BUFLEN, infile) == NULL) {
	fclose(infile);
	return CNV_EMPTY_FILE;
    }
    tokptr = buf;

    // Read in the file a token at a time
    while ((token = next_token()) != NULL) {

	switch (state) {

	    case SCANNING:

		// Capture a label which may be the DEF name
		if (token[strlen(token)-1] == ':') {
		    memset(lastDefName, 0, BUFLEN);
		    strncpy(lastDefName, token, strlen(token)-1);
		}

		// Got a ForwardDirection node?
		if (!strcmp(token, "ForwardDirection")) {
		    state = READING_FORWARDDIRECTION;
		    count = 0;
		}

		// Got an AmbientLight node?
		if (!strcmp(token, "AmbientLight")) {
		    state = READING_AMBIENTINTENSITY;
		    count = 0;
		}

		// Got a DirectionalLight node?
		if (!strcmp(token, "DirectionalLight")) {
		    state = READING_DIRECTIONALLIGHT;
		    count = 0;
		}

		// Got a CameraPlacement node?
		if (!strcmp(token, "CameraPlacement")) {
		    state = READING_CAMERAPLACEMENT;
		    count = 0;
		}

		// Got a TriMesh node?
		if (!strcmp(token, "TriMesh")) {
		    state = READING_TRIMESH_SPEC;
		    count = 0;
		}
		break;

	    case READING_FORWARDDIRECTION:

		switch (count++) {
		    case 0:
		    	// The first token is the '('
			break;
		    case 1:
			if ((sscanf(token, "%d", &direction[0])) < 1) {
			    fclose(infile);
			    return CNV_BAD_DIRECTION;
			}
			break;
		    case 2:
			if ((sscanf(token, "%d", &direction[1])) < 1) {
			    fclose(infile);
			    return CNV_BAD_DIRECTION;
			}
			break;
		    case 3:
			if ((sscanf(token, "%d", &direction[2])) < 1) {
			    fclose(infile);
			    return CNV_BAD_DIRECTION;
			}
			state = SCANNING;
			break;
		}
		break;

	    case READING_TRIMESH_SPEC:

		// Ignore the left paren
		if (count == 0) {
		    count++;
		    break;
		}
		if ((sscanf(token, "%d", &meshParms[count])) < 1) {
		    fclose(infile);
		    return CNV_BAD_TRI_MESH_SPEC;
		}
		if (++count > 6) {

		    // Create the Shape node
		    sp = newShape(lastDefName, NUM_POINTS, NUM_TRIANGLES);
		    if (sp == NULL) {
			fclose(infile);
			return CNV_CANT_ALLOC_NODE;
		    }

		    // Display the node and progress to date
		    memset(outputLine, 0, BUFLEN);
		    if (lastDefName[0] == '\0')
			sprintf(outputLine,
			    "(no name): %d triangles %d points %d edges",
			    NUM_TRIANGLES, NUM_POINTS, NUM_EDGES);
		    else
			sprintf(outputLine,
			    "%s:  %d triangles %d points %d edges",
			    lastDefName, NUM_TRIANGLES, NUM_POINTS, NUM_EDGES);
		    filePos = ftell(infile);
		    pct = 100 * ((float)filePos / fileLength)+0.5;
		    update_display(pct, outputLine);

		    state = READING_TRIANGLES;
		    count = 0;
		}
		break;

	    case READING_TRIANGLES:

		if ((sscanf(token, "%d", &index)) < 1) {
		    fclose(infile);
		    return CNV_BAD_TRIANGLE;
		}
		switch (count%3) {
		    case 0:
			sp->geometry.coordIndex.v[count/3].first = index;
			break;
		    case 1:
			// When the forward direction is -ve Z, we have to
			// swap the coordIndexes so the triangles will remain
			// counterclockwise
			if (direction[2] == -1)
			    sp->geometry.coordIndex.v[count/3].third = index;
			else
			    sp->geometry.coordIndex.v[count/3].second = index;
			break;
		    case 2:
			if (direction[2] == -1)
			    sp->geometry.coordIndex.v[count/3].second = index;
			else
			    sp->geometry.coordIndex.v[count/3].third = index;
			break;
		}
		if (++count >= NUM_TRIANGLES*3) {
		    state = READING_EDGES;
		    count = 0;
		}
		break;

	    case READING_EDGES:

		// Ignore the lines that define edges
		if (++count >= NUM_EDGES*4) {
		    state = READING_POINTS;
		    count = 0;
		}
		break;

	    case READING_POINTS:

		if ((sscanf(token, "%f", &coord)) < 1) {
		    fclose(infile);
		    return CNV_BAD_POINTS;
		}
		switch (count%3) {
		    case 0:
			sp->geometry.coord.point.v[count/3].x = coord;
			break;
		    case 1:
			sp->geometry.coord.point.v[count/3].y = coord;
			break;
		    case 2:
			// If forward direction is -ve Z, negate the Z value
			if (direction[2] == -1)
			    sp->geometry.coord.point.v[count/3].z = -coord;
			else
			    sp->geometry.coord.point.v[count/3].z = coord;
			break;
		}
		if (++count >= NUM_POINTS*3) {
		    state = SEARCHING_FOR_DIFFUSECOLOR;
		}
		break;

	    case SEARCHING_FOR_DIFFUSECOLOR:

		if (!strcmp(token, "DiffuseColor")) {
		    count = 0;
		    state = READING_DIFFUSECOLOR;
		}
		break;

	    case READING_DIFFUSECOLOR:

		switch (count++) {
		    case 0:
		    	// The first token is the '('
			break;
		    case 1:
			if ((sscanf(token, "%f", &color)) < 1) {
			    fclose(infile);
			    return CNV_BAD_COLOR;
			}
			sp->appearance.material.diffuseColor.r = color;
			break;
		    case 2:
			if ((sscanf(token, "%f", &color)) < 1) {
			    fclose(infile);
			    return CNV_BAD_COLOR;
			}
			sp->appearance.material.diffuseColor.g = color;
			break;
		    case 3:
			if ((sscanf(token, "%f", &color)) < 1) {
			    fclose(infile);
			    return CNV_BAD_COLOR;
			}
			sp->appearance.material.diffuseColor.b = color;
			state = SCANNING;
			memset(lastDefName, 0, BUFLEN);
			break;
		}
		break;

	    case READING_CAMERAPLACEMENT:

		switch (count++) {
		    case 0:
		    	// The first token is the '('
			break;
		    case 1:
			if ((sscanf(token, "%lf", &pos.x)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			break;
		    case 2:
			if ((sscanf(token, "%lf", &pos.y)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			break;
		    case 3:
			if ((sscanf(token, "%lf", &pos.z)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			if (direction[2] == -1)
			    pos.z = -pos.z;
			break;
		    case 4:
			if ((sscanf(token, "%lf", &lookat.x)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			break;
		    case 5:
			if ((sscanf(token, "%lf", &lookat.y)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			break;
		    case 6:
			if ((sscanf(token, "%lf", &lookat.z)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			if (direction[2] == -1)
			    lookat.z = -lookat.z;
			break;
		    case 7:
			if ((sscanf(token, "%lf", &up.x)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			break;
		    case 8:
			if ((sscanf(token, "%lf", &up.y)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			break;
		    case 9:
			if ((sscanf(token, "%lf", &up.z)) < 1) {
			    fclose(infile);
			    return CNV_BAD_VIEWPOINT;
			}
			if (direction[2] == -1)
			    up.z = -up.z;

			// Create the Viewpoint node
			vp = newViewpoint(lastDefName);
			if (vp == NULL) {
			    fclose(infile);
			    return CNV_CANT_ALLOC_NODE;
			}
			// Convert to VRML format
			Convert_Camera_Model(&pos, &lookat, &up, &axis, &angle);
			vp->orientation.x = (float)axis.x;
			vp->orientation.y = (float)axis.y;
			vp->orientation.z = (float)axis.z;
			vp->orientation.angle = (float)angle;
			vp->position.x = (float)pos.x;
			vp->position.y = (float)pos.y;
			vp->position.z = (float)pos.z;

			state = SCANNING;
			memset(lastDefName, 0, BUFLEN);
			break;
		}
		break;

	    case READING_DIRECTIONALLIGHT:

		switch (count++) {
		    case 0:
		    	// The first token is the '('
			break;
		    case 1:
			if ((sscanf(token, "%f", &lightDirection[0])) < 1) {
			    fclose(infile);
			    return CNV_BAD_LIGHT;
			}
			break;
		    case 2:
			if ((sscanf(token, "%f", &lightDirection[1])) < 1) {
			    fclose(infile);
			    return CNV_BAD_LIGHT;
			}
			break;
		    case 3:
			if ((sscanf(token, "%f", &lightDirection[2])) < 1) {
			    fclose(infile);
			    return CNV_BAD_LIGHT;
			}

			// Create the DirectionalLight node
			lp = newLight(lastDefName);
			if (lp == NULL) {
			    fclose(infile);
			    return CNV_CANT_ALLOC_NODE;
			}
			lp->ambientIntensity = ambientIntensity;
			lp->direction.x = lightDirection[0];
			lp->direction.y = lightDirection[1];
			if (direction[2] == -1)
			    lp->direction.z = -lightDirection[2];
			else
			    lp->direction.z = lightDirection[2];
			state = SEARCHING_FOR_LIGHT;
			memset(lastDefName, 0, BUFLEN);
			break;
		}
		break;

	    case SEARCHING_FOR_LIGHT:

		if (!strcmp(token, "Light")) {
		    state = READING_LIGHT;
		    count = 0;
		}
		break;

	    case READING_LIGHT:

		switch (count++) {
		    case 0:
		    	// The first token is the '('
			break;
		    case 1:
			// Then the on flag
			if (!strcmp(token, "True")) {
			    lp->on = 1;
			} else if (!strcmp(token, "False")) {
			    lp->on = 0;
			}
			break;
		    case 2:
			// Then intensity
			if ((sscanf(token, "%f", &intensity)) < 1) {
			    fclose(infile);
			    return CNV_BAD_LIGHT;
			}
			lp->intensity = intensity;
			break;
		    case 3:
			// Then three color values
			if ((sscanf(token, "%f", &color)) < 1) {
			    fclose(infile);
			    return CNV_BAD_LIGHT;
			}
			lp->color.r = color;
			break;
		    case 4:
			if ((sscanf(token, "%f", &color)) < 1) {
			    fclose(infile);
			    return CNV_BAD_LIGHT;
			}
			lp->color.g = color;
			break;
		    case 5:
			if ((sscanf(token, "%f", &color)) < 1) {
			    fclose(infile);
			    return CNV_BAD_LIGHT;
			}
			lp->color.b = color;

			// Create the DirectionalLight node
			state = SCANNING;
			break;
		}
		break;

	    case READING_AMBIENTINTENSITY:

		switch (count++) {
		    case 0:
		    case 1:
		    case 2:
		    case 3:
		    case 4:
		    	// Ignore ( ) Light ( True
			break;
		    case 5:
			// Intensity
			if ((sscanf(token, "%f", &ambientIntensity)) < 1) {
			    fclose(infile);
			    return CNV_BAD_LIGHT;
			}
			state = SCANNING;
			break;
		}
		break;
	}
    }

    fclose(infile);
    return CNV_OK;
}

// Convert from our internal format to VRML 97 and write the file
int write_out(char *fileName)
{
    FILE* outfile;
    struct Shape* sp;
    struct Viewpoint* vp;
    struct DirectionalLight *lp;
    int nNodes;
    int thisNode = 0;
    int i, j;

    nNodes = numberOfViewpoints() + numberOfLights() + numberOfShapes();

    if (!strcmp(fileName, "stdout")) {
	outfile = stdout;
    } else {
	if ((outfile = fopen(fileName, WRITEMODE)) == NULL)
	    return CNV_CANT_OPEN;
    }

    fprintf(outfile, "#VRML V2.0 utf8\n");

    fprintf(outfile, "NavigationInfo {\n");
    fprintf(outfile, "  type [ \"WALK\" \"ANY\" ]\n");
    // If we've got lights, turn off the headlight
    if (numberOfLights() > 0) {
	fprintf(outfile, "  headlight FALSE\n");
    }
    fprintf(outfile, "}\n");

    // Print viewpoints
    for (i=0; i<numberOfViewpoints(); i++) {
	vp = getViewpoint(i);
	if (viewpointIsUnique(i)) {
	    if (vp->defName != NULL)
		fprintf(outfile, "DEF %s ", vp->defName);
	    fprintf(outfile, "Viewpoint{\n");
	    fprintf(outfile, "  position %f %f %f\n",
		vp->position.x, vp->position.y, vp->position.z);
	    fprintf(outfile, "  orientation %f %f %f %f\n",
		vp->orientation.x, vp->orientation.y, vp->orientation.z,
		vp->orientation.angle);
	    fprintf(outfile, "  description \"%s\"\n",
		vp->description);
	    fprintf(outfile, "}\n");
	}
	update_display(100*thisNode++/nNodes, "");
    }

    // Directional lights and shapes go inside a transorm node
    fprintf(outfile, "Transform {\n");
    fprintf(outfile, "  children [\n");

    for (i=0; i<numberOfLights(); i++) {
	lp = getLight(i);
	if (lightIsUnique(i)) {
	    if (lp->defName != NULL)
		fprintf(outfile, "    DEF %s ", lp->defName);
	    fprintf(outfile, "DirectionalLight {\n");
	    fprintf(outfile, "      color %f %f %f\n",
		lp->color.r, lp->color.g, lp->color.b);
	    fprintf(outfile, "      intensity %f\n", lp->intensity);
	    fprintf(outfile, "      ambientIntensity %f\n",
		lp->ambientIntensity);
	    fprintf(outfile, "      direction %f %f %f\n",
		lp->direction.x, lp->direction.y, lp->direction.z);
	    fprintf(outfile, "    }\n");
	}
	update_display(100*thisNode++/nNodes, "");
    }

    for (i=0; i<numberOfShapes(); i++) {
	sp = getShape(i);
	if (sp->defName != NULL)
	    fprintf(outfile, "    DEF %s ", sp->defName);
	fprintf(outfile, "Shape {\n");
	fprintf(outfile, "      appearance Appearance {\n");
	fprintf(outfile, "        material Material {\n");
	fprintf(outfile, "          diffuseColor %f %f %f\n",
	    sp->appearance.material.diffuseColor.r,
	    sp->appearance.material.diffuseColor.g,
	    sp->appearance.material.diffuseColor.b);
	fprintf(outfile, "        }\n");
	fprintf(outfile, "      }\n");
	fprintf(outfile, "      geometry IndexedFaceSet {\n");
	fprintf(outfile, "        coord Coordinate {\n");
	fprintf(outfile, "          point [\n");
	for (j=0; j<sp->geometry.coord.point.numPoints; j++) {
	    fprintf(outfile, "            %f %f %f\n",
		sp->geometry.coord.point.v[j].x,
		sp->geometry.coord.point.v[j].y,
		sp->geometry.coord.point.v[j].z);
	}
	fprintf(outfile, "          ]\n");
	fprintf(outfile, "        }\n");
	fprintf(outfile, "        coordIndex [\n");
	for (j=0; j<sp->geometry.coordIndex.numPoints; j++) {
	    fprintf(outfile, "          %d %d %d -1\n",
		sp->geometry.coordIndex.v[j].first,
		sp->geometry.coordIndex.v[j].second,
		sp->geometry.coordIndex.v[j].third);
	}
	fprintf(outfile, "        ]\n");
	fprintf(outfile, "      }\n");
	fprintf(outfile, "    }\n");
	update_display(100*thisNode++/nNodes, "");
    }
    fprintf(outfile, "  ]\n");
    fprintf(outfile, "}\n");
    fclose(outfile);
    return CNV_OK;
}

// Do any cleanup necessary
void clean_up()
{
    newSceneGraph();
}
