//
// vrml.c -- Functions to create, destroy, fetch VRML nodes
//
// These functions do not assume anything about the platform, and should be
// usable on both the Windows version and the console version.
//
// Copyright (C) 2000 by Bob Crispen <crispen@hiwaay.net> doing business as
//   Foxy Software
//
// LICENSE:
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <malloc.h>
#include <string.h>
#include "vrml.h"

int nViewpoints = 0;
struct Viewpoint** v;
int nShapes = 0;
struct Shape** s;
int nLights = 0;
struct DirectionalLight** l;

// Return the number of shape nodes in our scene graph
int numberOfShapes()
{
    return nShapes;
}

// Get the pointer to a given shape
struct Shape* getShape(int i)
{
    return s[i];
}

// Create a new Shape node and return a pointer to it
struct Shape* newShape(char* defName, int numPoints, int numTriangles)
{
    // Allocate room in our scene graph array
    if (nShapes == 0) {
	s = (struct Shape**)malloc(sizeof(struct Shape*));
    } else {
	s = (struct Shape**)realloc(s, sizeof(struct Shape*) * (nShapes+1));
    }

    // Make a new Shape node
    if (s == NULL) return NULL;
    s[nShapes] = (struct Shape*)malloc(sizeof(struct Shape));
    if (s[nShapes] == NULL) return NULL;

    // Copy in the name if there is one
    if (defName[0] == '\0') {
	s[nShapes]->defName = NULL;
    } else {
	s[nShapes]->defName = (char*)malloc(strlen(defName)+1);
	if (s[nShapes]->defName == NULL) return NULL;
	strcpy(s[nShapes]->defName, defName);
    }

    // Allocate space for our points
    s[nShapes]->geometry.coord.point.numPoints = numPoints;
    s[nShapes]->geometry.coord.point.v =
	(struct SFVec3f*)malloc(numPoints * sizeof(struct SFVec3f));
    if (s[nShapes]->geometry.coord.point.v == NULL) return NULL;

    // Allocate space for our triangles
    s[nShapes]->geometry.coordIndex.numPoints = numTriangles;
    s[nShapes]->geometry.coordIndex.v =
	(struct TriangleIndex*)
	malloc(numTriangles * sizeof(struct TriangleIndex));
    if (s[nShapes]->geometry.coordIndex.v == NULL) return NULL;

    // XXX Fill in some other fields

    return s[nShapes++];
}

// Return the number of viewpoint nodes in our scene graph
int numberOfViewpoints()
{
    return nViewpoints;
}

// Get the pointer to a given viewpoint
struct Viewpoint* getViewpoint(int i)
{
    return v[i];
}

// They've got multiple identical viewpoints in some worlds.  Don't duplicate
// an existing viewpoint.
int viewpointIsUnique(int thisViewpoint)
{
    int i;

    if (thisViewpoint==0)
	return 1;
    for (i=0; i<thisViewpoint; i++) {
	if (v[i]->fieldOfView != v[thisViewpoint]->fieldOfView)
	    return 1;
	if (v[i]->jump != v[thisViewpoint]->jump)
	    return 1;
	if (v[i]->orientation.x != v[thisViewpoint]->orientation.x)
	    return 1;
	if (v[i]->orientation.y != v[thisViewpoint]->orientation.y)
	    return 1;
	if (v[i]->orientation.z != v[thisViewpoint]->orientation.z)
	    return 1;
	if (v[i]->orientation.angle != v[thisViewpoint]->orientation.angle)
	    return 1;
	if (v[i]->position.x != v[thisViewpoint]->position.x)
	    return 1;
	if (v[i]->position.y != v[thisViewpoint]->position.y)
	    return 1;
	if (v[i]->position.z != v[thisViewpoint]->position.z)
	    return 1;
    }
    return 0;
}

// Create a new viewpoint node and return a pointer to it
struct Viewpoint* newViewpoint(char* defName)
{
    // Allocate room in our scene graph array
    if (nViewpoints == 0) {
	v = (struct Viewpoint**)malloc(sizeof(struct Viewpoint*));
    } else {
	v = (struct Viewpoint**)
	    realloc(v, sizeof(struct Viewpoint*) * (nViewpoints+1));
    }

    // Make a new Viewpoint node
    if (v == NULL) return NULL;
    v[nViewpoints] = (struct Viewpoint*)malloc(sizeof(struct Viewpoint));
    if (v[nViewpoints] == NULL) return NULL;

    // Copy in the name if there is one
    if (defName[0] == '\0') {
	v[nViewpoints]->defName = NULL;
    } else {
	v[nViewpoints]->defName = (char*)malloc(strlen(defName)+1);
	if (v[nViewpoints]->defName == NULL) return NULL;
	strcpy(v[nViewpoints]->defName, defName);

	// XXX For this program we'll use the DEF name for the description
	v[nViewpoints]->description = (char*)malloc(strlen(defName)+1);
	if (v[nViewpoints]->description == NULL) return NULL;
	strcpy(v[nViewpoints]->description, defName);
    }
    // Initialize fields to default VRML spec values
    v[nViewpoints]->fieldOfView = 0.785398;
    v[nViewpoints]->jump = 1;
    v[nViewpoints]->orientation.x = 0.0;
    v[nViewpoints]->orientation.y = 0.0;
    v[nViewpoints]->orientation.z = 1.0;
    v[nViewpoints]->orientation.angle = 0.0;
    v[nViewpoints]->position.x = 0.0;
    v[nViewpoints]->position.y = 0.0;
    v[nViewpoints]->position.z = 10.0;

    return v[nViewpoints++];
}

// Return the number of DirectionalLight nodes in our scene graph
int numberOfLights()
{
    return nLights;
}

// Get the pointer to a given DirectionalLight
struct DirectionalLight* getLight(int i)
{
    return l[i];
}

// They've got multiple identical lights in some worlds.  Don't duplicate
// an existing light.
int lightIsUnique(int thisLight)
{
    int i;
    if (thisLight==0)
	return 1;
    for (i=0; i<thisLight; i++) {
	if (l[i]->ambientIntensity != l[thisLight]->ambientIntensity)
	    return 1;
	if (l[i]->color.r != l[thisLight]->color.r)
	    return 1;
	if (l[i]->color.g != l[thisLight]->color.g)
	    return 1;
	if (l[i]->color.b != l[thisLight]->color.b)
	    return 1;
	if (l[i]->direction.x != l[thisLight]->direction.x)
	    return 1;
	if (l[i]->direction.y != l[thisLight]->direction.y)
	    return 1;
	if (l[i]->direction.z != l[thisLight]->direction.z)
	    return 1;
	if (l[i]->intensity != l[thisLight]->intensity)
	    return 1;
	if (l[i]->on != l[thisLight]->on)
	    return 1;
    }
    return 0;
}

// Create a new viewpoint node and return a pointer to it
struct DirectionalLight* newLight(char* defName)
{
    // Allocate room in our scene graph array
    if (nLights == 0) {
	l = (struct DirectionalLight**)malloc(sizeof(struct DirectionalLight*));
    } else {
	l = (struct DirectionalLight**)
	    realloc(l, sizeof(struct DirectionalLight*) * (nLights+1));
    }

    // Make a new DirectionalLight node
    if (l == NULL) return NULL;
    l[nLights] = (struct DirectionalLight*)
	malloc(sizeof(struct DirectionalLight));
    if (l[nLights] == NULL) return NULL;

    // Copy in the name if there is one
    if (defName[0] == '\0') {
	l[nLights]->defName = NULL;
    } else {
	l[nLights]->defName = (char*)malloc(strlen(defName)+1);
	if (l[nLights]->defName == NULL) return NULL;
	strcpy(l[nLights]->defName, defName);
    }
    // Initialize fields to default VRML spec values
    l[nLights]->ambientIntensity = 0.0;
    l[nLights]->color.r = 1;
    l[nLights]->color.g = 1;
    l[nLights]->color.b = 1;
    l[nLights]->direction.x = 0.0;
    l[nLights]->direction.y = 0.0;
    l[nLights]->direction.z = -1.0;
    l[nLights]->intensity = 1.0;
    l[nLights]->on = 1;

    return l[nLights++];
}

// Free up all the mallocs in our scene graph
void newSceneGraph()
{
    int i;
    for (i=nShapes-1; i>=0; --i) {
	free(s[i]->geometry.coordIndex.v);
	free(s[i]->geometry.coord.point.v);
	free(s[i]->defName);
	free(s[i]);
    }
    nShapes = 0;

    for (i=nViewpoints-1; i>=0; --i) {
	free(v[i]->defName);
	free(v[i]->description);
	free(v[i]);
    }
    nViewpoints = 0;

    for (i=nLights-1; i>=0; --i) {
	free(l[i]->defName);
	free(l[i]);
    }
    nLights = 0;
}
